from __future__ import annotations
import asyncio
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Optional, Union

from gallia.db.db_handler import DBHandler, LogMode
from gallia.penlab import PowerSupply
from gallia.uds.core import service
from gallia.uds.core.client import UDSClient, UDSRequestConfig
from gallia.penlog import Logger
from gallia.uds.core.constants import DataIdentifier
from gallia.uds.core.exception import ResponseException
from gallia.uds.helpers import as_exception, raise_for_error
from gallia.uds.core.utils import from_bytes
from gallia.transports.base import BaseTransport


@dataclass
class ECUState:
    session: int = 1
    security_access_level: Optional[int] = None


class ECU(UDSClient):
    """ECU is a high level interface wrapping a UDSClient class. It provides
    semantically correct higher level interfaces such as read_session()
    or ping(). Vendor specific implementations can be derived from this
    class. For the arguments of the constructor, please check uds.uds.UDS.
    """

    def __init__(
        self,
        transport: BaseTransport,
        timeout: Optional[float] = None,
        max_retry: int = 1,
        power_supply: Optional[PowerSupply] = None,
    ) -> None:

        super().__init__(transport, timeout, max_retry)
        self.logger = Logger(component="ecu", flush=True)
        self.power_supply = power_supply
        self.state = ECUState()
        self.db_handler: Optional[DBHandler] = None

    async def connect(self) -> None:
        ...

    async def properties(
        self, fresh: bool = False, config: Optional[UDSRequestConfig] = None
    ) -> dict:
        return {}

    async def ping(
        self, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.TesterPresentResponse]:
        """Send an UDS TesterPresent message.

        Returns:
            UDS response.
        """
        return await self.tester_present(suppress_response=False, config=config)

    async def read_session(self, config: Optional[UDSRequestConfig] = None) -> int:
        """Read out current session.

        Returns:
            The current session as int.
        """
        resp = await self.read_data_by_identifier(
            DataIdentifier.ActiveDiagnosticSessionDataIdentifier, config=config
        )
        if isinstance(resp, service.NegativeResponse):
            raise as_exception(resp)
        return from_bytes(resp.data_record)

    async def set_session_pre(
        self, level: int, config: Optional[UDSRequestConfig] = None
    ) -> bool:
        """set_session_pre() is called before the diagnostic session control
        pdu is written on the wire. Implement this if there are special
        preconditions for a particular session, such as disabling error
        logging.

        Args:
            uds: The UDSClient class where this hook is embedded. The caller typically
                 calls this function with `self` as the first argument.
            session: The desired session identifier.
        Returns:
            True on success, False on error.
        """

    async def set_session_post(
        self, level: int, config: Optional[UDSRequestConfig] = None
    ) -> bool:
        """set_session_post() is called after the diagnostic session control
        pdu was written on the wire. Implement this if there are special
        cleanup routines or sleeping until a certain moment is required.

        Args:
            uds: The UDSClient class where this hook is embedded. The caller typically
                 calls this function with `self` as the first argument.
            session: The desired session identifier.
        Returns:
            True on success, False on error.
        """

    async def power_cycle(self, sleep: int = 5) -> bool:
        if self.power_supply is None:
            self.logger.log_warning("no power_supply available")
            return False

        await self.power_supply.power_cycle(sleep, self.wait_for_ecu)
        return True

    async def leave_session(
        self, level: int, config: Optional[UDSRequestConfig] = None
    ) -> bool:
        """leave_session() is a hook which can be called explicitly by a
        scanner when a session is to be disabled. Use this hook if resetting
        the ECU is required, e.g. when disabling the programming session.

        Args:
            uds: The UDSClient class where this hook is embedded. The caller typically
                 calls this function with `self` as the first argument.
            session: The desired session identifier.
        Returns:
            True on success, False on error.
        """
        resp = await self.ecu_reset(0x01)
        if isinstance(resp, service.NegativeResponse):
            await self.power_cycle()
        else:
            await self.wait_for_ecu()

        resp = await self.set_session(0x01, config=config)
        if isinstance(resp, service.NegativeResponse):
            return await self.power_cycle()
        return True

    async def set_session(
        self, level: int, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.DiagnosticSessionControlResponse]:
        config = config if config is not None else UDSRequestConfig()

        if not config.skip_hooks:
            await self.set_session_pre(level, config=config)

        resp = await self.diagnostic_session_control(level, config=config)

        if not isinstance(resp, service.NegativeResponse) and not config.skip_hooks:
            await self.set_session_post(level, config=config)

        return resp

    async def read_dtc(
        self, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.ReportDTCByStatusMaskResponse]:
        """Read all dtc records from the ecu."""
        return await self.read_dtc_information_report_dtc_by_status_mask(
            0xFF, config=config
        )

    async def clear_dtc(
        self, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.ClearDiagnosticInformationResponse]:
        """Clear all dtc records on the ecu."""
        return await self.clear_diagnostic_information(0xFFFFFF, config=config)

    async def read_vin(
        self, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.ReadDataByIdentifierResponse]:
        """Read the VIN of the vehicle"""
        return await self.read_data_by_identifier(0xF190, config=config)

    async def transmit_data(
        self,
        data: bytes,
        block_length: int,
        max_block_length: int = 0xFFF,
        config: Optional[UDSRequestConfig] = None,
    ) -> None:
        """transmit_data splits the data to be sent in several blocks of size block_length,
        transfers all of them and concludes the transmission with RequestTransferExit"""
        if block_length > max_block_length:
            self.logger.log_warning(f"Limiting block size to 0x{max_block_length:x}")
            block_length = max_block_length
        # block_length includes the service identifier and block counter; payload must be smaller
        payload_size = block_length - 2
        counter = 0
        for i in range(0, len(data), payload_size):
            counter += 1
            payload = data[i : i + payload_size]
            self.logger.log_debug(
                f"Transferring block 0x{counter:02x} "
                f"with payload size 0x{len(payload):x}"
            )
            resp: service.UDSResponse = await self.transfer_data(
                counter & 0xFF, payload, config=config
            )
            raise_for_error(resp, f"Transmitting data failed at index 0x{i:x}")
        resp = await self.request_transfer_exit(config=config)
        raise_for_error(resp)

    async def wait_for_ecu(self, timeout: float = 60) -> bool:
        """wait for ecu to be alive again (eg. after reset)
        Wait at most timeout"""
        ret = False
        try:
            await asyncio.wait_for(self._wait_for_ecu(), timeout=timeout)
            ret = True
        except asyncio.TimeoutError:
            self.logger.log_critical("Timeout while waiting for ECU!")
        return ret

    async def _wait_for_ecu(self) -> None:
        """wait for ecu to be alive again (eg. after reset)
        Internal method without timeout"""
        self.logger.log_info("waiting for ECU…")
        while True:
            try:
                await asyncio.sleep(1)
                await self.reconnect()
                await self.ping()
                break
            # If the network is down or anything else is broken,
            # then an OSError is raised. Raise this kind of errors.
            except OSError:
                raise
            except Exception as e:
                self.logger.log_debug(f"ECU not ready: {e.__class__.__name__} {e}")
        self.logger.log_info("ECU ready")

    async def update_state(
        self, request: service.UDSRequest, response: service.UDSResponse
    ) -> None:
        if isinstance(response, service.DiagnosticSessionControlResponse):
            self.state.session = response.diagnostic_session_type

        if (
            isinstance(response, service.ReadDataByIdentifierResponse)
            and response.data_identifier
            == DataIdentifier.ActiveDiagnosticSessionDataIdentifier
        ):
            self.state.session = int.from_bytes(response.data_record, "big")

        if isinstance(response, service.SecurityAccessResponse):
            self.state.security_access_level = response.security_access_type - 1

        if isinstance(response, service.ECUResetResponse):
            self.state = ECUState()

    async def _request(
        self, request: service.UDSRequest, config: Optional[UDSRequestConfig] = None
    ) -> service.UDSResponse:
        """Sends a raw UDS request and returns the response.
        Network errors are handled via exponential backoff.
        Pending errors, triggered by the ECU are resolved as well.

        :param request: request to send
        :param config: The request config parameters
        :return: The response.
        """
        response = None
        exception: Optional[Exception] = None
        send_time = datetime.now(timezone.utc).astimezone()
        receive_time = None

        try:
            response = await super()._request(request, config)
            receive_time = datetime.now(timezone.utc).astimezone()
            return response
        except ResponseException as e:
            exception = e
            response = e.response
            raise
        except Exception as e:
            exception = e
            raise
        finally:
            try:
                if self.db_handler is not None:
                    log_mode = LogMode.implicit

                    if (
                        config is not None
                        and config.tags is not None
                        and "ANALYZE" in config.tags
                    ):
                        log_mode = LogMode.emphasized

                    await self.db_handler.insert_scan_result(
                        self.state.__dict__,
                        service.UDSRequest.parse_dynamic(request.pdu),
                        response,
                        exception,
                        send_time,
                        receive_time,
                        log_mode,
                    )
            except Exception as e:
                self.logger.log_warning(f"Could not log messages to database: {e}")

            if response is not None:
                await self.update_state(request, response)



from __future__ import annotations

import asyncio
import binascii
from typing import Optional, TypedDict

from gallia.transports.base import BaseTransport, TargetURI


_TCP_SPEC_TYPE = TypedDict("_TCP_SPEC_TYPE", {})
tcp_spec: dict = {}
assertion_str = "bug: transport is not connected"


class TCPTransport(BaseTransport, scheme="tcp", spec=tcp_spec):
    def __init__(
        self,
        target: TargetURI,
        reader: Optional[asyncio.StreamReader] = None,
        writer: Optional[asyncio.StreamWriter] = None,
    ) -> None:
        super().__init__(target)
        self.reader = reader
        self.writer = writer

    async def connect(self, timeout: Optional[float] = None) -> None:
        assert (
            self.reader is None and self.writer is None
        ), "bug: transport is already connected"
        self.reader, self.writer = await asyncio.wait_for(
            asyncio.open_connection(self.target.hostname, self.target.port), timeout
        )

    async def reconnect(self, timeout: Optional[float] = None) -> None:
        await self.terminate()
        await self.connect(timeout)

    async def terminate(self) -> None:
        assert self.reader is not None and self.writer is not None, assertion_str

        self.writer.close()
        await self.writer.wait_closed()
        self.reader = None
        self.writer = None

    async def write(
        self,
        data: bytes,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> int:
        assert self.reader is not None and self.writer is not None, assertion_str

        self.logger.log_write(data.hex(), tags=tags)
        self.writer.write(data)
        await asyncio.wait_for(self.writer.drain(), timeout)
        return len(data)

    async def read(
        self,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> bytes:
        assert self.reader is not None and self.writer is not None, assertion_str

        data = await asyncio.wait_for(self.reader.read(self.BUFSIZE), timeout)
        self.logger.log_read(data.hex(), tags=tags)
        return data

    async def sendto(
        self,
        data: bytes,
        dst: int,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> int:
        raise RuntimeError("sendto() is not implemented")

    async def recvfrom(
        self,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> tuple[int, bytes]:
        raise RuntimeError("recvfrom() is not implemented")


class TCPLineSepTransport(TCPTransport, scheme="tcp-lines", spec=tcp_spec):
    async def write(
        self,
        data: bytes,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> int:
        assert self.reader is not None and self.writer is not None, assertion_str

        d = binascii.hexlify(data)
        self.logger.log_write(data.hex(), tags=tags)
        self.writer.write(d + b"\n")
        await asyncio.wait_for(self.writer.drain(), timeout)
        return len(data)

    async def read(
        self,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> bytes:
        assert self.reader is not None and self.writer is not None, assertion_str

        data = await asyncio.wait_for(self.reader.readline(), timeout)
        d = data.decode().strip()
        self.logger.log_read(d, tags=tags)
        return binascii.unhexlify(d)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gallia',
 'gallia.cursed_hr',
 'gallia.db',
 'gallia.services',
 'gallia.transports',
 'gallia.uds',
 'gallia.uds.core',
 'gallia.udscan',
 'gallia.udscan.scanner']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=0.8.0,<0.9.0',
 'aiosqlite>=0.17.0,<0.18.0',
 'argcomplete>=2.0.0,<3.0.0',
 'pyopennetzteil>=0.1.2,<0.2.0',
 'pypenlog>=0.3.2,<0.4.0',
 'python-can>=4.0.0,<5.0.0',
 'pyxcp>=0.16.0,<0.17.0',
 'tabulate>=0.8.9,<0.9.0',
 'types-tabulate>=0.8.8,<0.9.0',
 'zstandard>=0.17.0,<0.18.0']

entry_points = \
{'console_scripts': ['cursed-hr = gallia.cursed_hr.cursed_hr:main',
                     'discover-xcp = gallia.udscan.scanner.find_xcp:main',
                     'gallia = gallia.cli:main'],
 'gallia_scanners': ['discover-can-ids = '
                     'gallia.udscan.scanner.find_can_ids:FindCanIDsScanner',
                     'discover-endpoints = '
                     'gallia.udscan.scanner.find_endpoints:FindEndpoints',
                     'discover-iso-tp-addr = '
                     'gallia.udscan.scanner.find_iso_tp_addr:FindISOTPAddrScanner',
                     'scan-dump-seeds = '
                     'gallia.udscan.scanner.scan_sa_dump_seeds:SaDumpSeeds',
                     'scan-identifiers = '
                     'gallia.udscan.scanner.scan_identifiers:ScanIdentifiers',
                     'scan-memory-functions = '
                     'gallia.udscan.scanner.scan_memory_functions:ScanWriteDataByAddress',
                     'scan-reset = gallia.udscan.scanner.scan_reset:ScanReset',
                     'scan-services = '
                     'gallia.udscan.scanner.scan_services:ScanServices',
                     'scan-sessions = '
                     'gallia.udscan.scanner.scan_sessions:IterateSessions',
                     'simple-dtc = gallia.udscan.scanner.simple_dtc:DTCScanner',
                     'simple-ecu-reset = '
                     'gallia.udscan.scanner.simple_ecu_reset:EcuReset',
                     'simple-get-vin = '
                     'gallia.udscan.scanner.simple_get_vin:GetVin',
                     'simple-iocbi = gallia.udscan.scanner.simple_iocbi:IOCBI',
                     'simple-ping = gallia.udscan.scanner.simple_ping:Ping',
                     'simple-read-by-identifier = '
                     'gallia.udscan.scanner.simple_read_by_identifier:ReadByIdentifier',
                     'simple-read-error-log = '
                     'gallia.udscan.scanner.simple_read_error_log:ReadErrorLog',
                     'simple-rmba = '
                     'gallia.udscan.scanner.simple_rmba:ReadMemoryByAddressScanner',
                     'simple-rtcl = gallia.udscan.scanner.simple_rtcl:RTCL',
                     'simple-send-pdu = '
                     'gallia.udscan.scanner.simple_send_pdu:SendPDU',
                     'simple-test-xcp = '
                     'gallia.udscan.scanner.simple_test_xcp:TestXCP',
                     'simple-wmba = '
                     'gallia.udscan.scanner.simple_wmba:WriteMemoryByAddressScanner',
                     'simple-write-by-identifier = '
                     'gallia.udscan.scanner.simple_write_by_identifier:WriteByIdentifier']}

setup_kwargs = {
    'name': 'gallia',
    'version': '1.0.0a4',
    'description': 'Extendable Pentesting Framework',
    'long_description': '# Gallia\n\n## Documentation\n\nhttps://fraunhofer-aisec.github.io/gallia\n',
    'author': 'AISEC Pentesting Team',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Fraunhofer-AISEC/gallia',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



import asyncio
import binascii
import sys
from argparse import Namespace
from typing import Union

from gallia.uds.core.service import NegativeResponse, _RoutineControlResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int, check_and_set_session


class RTCL(UDSScanner):
    """Start or stop a provided routine or request its results"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

        self.parser.add_argument(
            "--session",
            type=auto_int,
            default=0x01,
            help="The session in which the requests are made",
        )
        self.parser.add_argument(
            "routine_identifier", type=auto_int, help="The routine identifier"
        )
        self.parser.add_argument(
            "--start",
            action="store_true",
            help="Start the routine with a startRoutine request (this task is always executed first)",
        )
        self.parser.add_argument(
            "--stop",
            action="store_true",
            help="Stop the routine with a stopRoutine request "
            "(this task is executed after starting the routine if --start is given as well)",
        )
        self.parser.add_argument(
            "--results",
            action="store_true",
            help="Read the routine results with a requestRoutineResults request (this task is always executed last)",
        )
        self.parser.add_argument(
            "--start-parameters",
            metavar="HEXSTRING",
            type=binascii.unhexlify,
            default=b"",
            help="The routineControlOptionRecord passed to the startRoutine request",
        )
        self.parser.add_argument(
            "--stop-parameters",
            metavar="HEXSTRING",
            type=binascii.unhexlify,
            default=b"",
            help="The routineControlOptionRecord passed to the stopRoutine request",
        )
        self.parser.add_argument(
            "--results-parameters",
            metavar="HEXSTRING",
            type=binascii.unhexlify,
            default=b"",
            help="The routineControlOptionRecord passed to the stopRoutine request",
        )
        self.parser.add_argument(
            "--stop-delay",
            metavar="SECONDS",
            type=float,
            default=0.0,
            help="Delay the stopRoutine request by the given amount of seconds",
        )
        self.parser.add_argument(
            "--results-delay",
            metavar="SECONDS",
            type=float,
            default=0.0,
            help="Delay the requestRoutineResults request by the given amount of seconds",
        )

    async def main(self, args: Namespace) -> None:
        try:
            await check_and_set_session(self.ecu, args.session)
        except Exception as e:
            self.logger.log_critical(
                f"Could not change to session: 0x{args.session:02x}: {e.__class__.__name__} {e}"
            )
            sys.exit(1)

        if args.start is False and args.stop is False and args.results is False:
            self.logger.log_warning("No instructions were given (start/stop/results)")

        if args.start:
            resp: Union[
                NegativeResponse, _RoutineControlResponse
            ] = await self.ecu.routine_control_start_routine(
                args.routine_identifier, args.start_parameters
            )

            if isinstance(resp, NegativeResponse):
                self.logger.log_error(f"start_routine: {resp}")
            else:
                self.logger.log_summary("[start] Positive response:")
                self.logger.log_summary(f"hex: {resp.routine_status_record.hex()}")
                self.logger.log_summary(f"raw: {repr(resp.routine_status_record)}")

        if args.stop:
            delay = args.stop_delay

            if delay > 0:
                self.logger.log_info(
                    f"Delaying the request for stopping the routine by {delay} seconds"
                )
                await asyncio.sleep(delay)

            resp = await self.ecu.routine_control_stop_routine(
                args.routine_identifier, args.stop_parameters
            )

            if isinstance(resp, NegativeResponse):
                self.logger.log_error(f"stop routine: {resp}")
            else:
                self.logger.log_summary("[stop] Positive response:")
                self.logger.log_summary(f"hex: {resp.routine_status_record.hex()}")
                self.logger.log_summary(f"raw: {repr(resp.routine_status_record)}")

        if args.results:
            delay = args.results_delay

            if delay > 0:
                self.logger.log_info(
                    f"Delaying the request for the routine results by {delay} seconds"
                )
                await asyncio.sleep(delay)

            resp = await self.ecu.routine_control_request_routine_results(
                args.routine_identifier, args.results_parameters
            )

            if isinstance(resp, NegativeResponse):
                self.logger.log_error(f"request_routine_results: {resp}")
            else:
                self.logger.log_summary("[get result] Positive response:")
                self.logger.log_summary(f"hex: {resp.routine_status_record.hex()}")
                self.logger.log_summary(f"raw: {repr(resp.routine_status_record)}")



# PYTHON_ARGCOMPLETE_OK

import argparse
import os
import sys
from importlib.metadata import entry_points

import argcomplete


def main() -> None:
    parser = argparse.ArgumentParser()
    sp = parser.add_subparsers(
        metavar="CHOICE",
    )

    all_entries = entry_points()

    if "_ARGCOMPLETE" in os.environ:
        comp_line = os.environ["COMP_LINE"]
        comp_point = int(os.environ["COMP_POINT"])

        # For completion only the names are required, which results in a noticeably faster response
        # compared to the full loading process, which is necessary for the help
        for group in ("gallia_scanners", "gallia_scripts"):
            for entry in all_entries[group]:
                _parser = sp.add_parser(entry.name)

        if len(comp_line.lstrip().split(" ")) < 3:
            argcomplete.autocomplete(parser)
        else:
            entry_prefix_length = comp_line.find(" ") + 1

            os.environ["COMP_LINE"] = comp_line[entry_prefix_length:]
            os.environ["COMP_POINT"] = str(comp_point - entry_prefix_length)

            chosen_entry = comp_line.split()[1]

            try:
                entry_point = next(
                    i.load()
                    for i in all_entries["gallia_scanners"]
                    if i.name == chosen_entry
                )
                sys.exit(entry_point().run())
            except StopIteration:
                try:
                    entry_point = next(
                        i.load()
                        for i in all_entries["gallia_scripts"]
                        if i.name == chosen_entry
                    )
                    sys.exit(entry_point())
                except StopIteration:
                    pass
    else:
        if "gallia_scanners" in all_entries:
            for entry in all_entries["gallia_scanners"]:
                scanner_class = entry.load()

                _parser = sp.add_parser(
                    entry.name,
                    help=scanner_class.__doc__,
                )
                _parser.set_defaults(func=scanner_class().run)

        if "gallia_scripts" in all_entries:
            for entry in all_entries["gallia_scripts"]:
                entry_point = entry.load()

                _parser = sp.add_parser(
                    entry.name,
                    help=entry_point.__doc__,
                )
                _parser.set_defaults(func=entry_point)

        # Only pass a single argument to the args parser, otherwise it gets confused
        # with arguments belonging to one of the subparsers.
        if len(sys.argv) < 2:
            parser.print_help()
            parser.exit()

        args = parser.parse_args([sys.argv[1]])

        # Combine the first two arguments for subsequent parsers to work correctly
        sys.argv[0] += f" {sys.argv[1]}"
        sys.argv.pop(1)

        if not hasattr(args, "func"):
            parser.print_help()
            parser.exit()
        sys.exit(args.func())



from enum import IntEnum


class UDSIsoServices(IntEnum):
    ShowCurrentData = 0x01
    ShowFreezeFrameData = 0x02
    ShowstoredDiagnosticTroubleCodes = 0x03
    ClearDiagnosticTroubleCodesAndStoredValues = 0x04
    TestResultsOxygenSensorMonitoring = 0x05
    TestResultsOtherComponentSystemMonitoring = 0x06
    ShowPendingDiagnosticTroubleCodes = 0x07
    ControlOperationOfOnBoardComponentSystem = 0x08
    RequestVehicleInformation = 0x09
    PermanentDiagnosticTroubleCodes = 0x0A
    DiagnosticSessionControl = 0x10
    EcuReset = 0x11
    SecurityAccess = 0x27
    CommunicationControl = 0x28
    TesterPresent = 0x3E
    Authentication = 0x29
    AccessTimingParameter = 0x83
    SecuredDataTransmission = 0x84
    ControlDTCSetting = 0x85
    ResponseOnEvent = 0x86
    LinkControl = 0x87
    ReadDataByIdentifier = 0x22
    ReadMemoryByAddress = 0x23
    ReadScalingDataByIdentifier = 0x24
    ReadDataByPeriodicIdentifier = 0x2A
    DynamicallyDefineDataIdentifier = 0x2C
    WriteDataByIdentifier = 0x2E
    WriteMemoryByAddress = 0x3D
    ClearDiagnosticInformation = 0x14
    ReadDTCInformation = 0x19
    InputOutputControlByIdentifier = 0x2F
    RoutineControl = 0x31
    RequestDownload = 0x34
    RequestUpload = 0x35
    TransferData = 0x36
    RequestTransferExit = 0x37
    RequestFileTransfer = 0x38
    NegativeResponse = 0x7F


class UDSErrorCodes(IntEnum):
    generalReject = 0x10
    serviceNotSupported = 0x11
    subFunctionNotSupported = 0x12
    incorrectMessageLengthOrInvalidFormat = 0x13
    responseTooLong = 0x14
    busyRepeatRequest = 0x21
    conditionsNotCorrect = 0x22
    requestSequenceError = 0x24
    noResponseFromSubnetComponent = 0x25
    failurePreventsExecutionOfRequestedAction = 0x25
    requestOutOfRange = 0x31
    securityAccessDenied = 0x33
    authenticationRequired = 0x34
    invalidKey = 0x35
    exceededNumberOfAttempts = 0x36
    requiredTimeDelayNotExpired = 0x37
    secureDataTransmissionRequired = 0x38
    secureDataTransmissionNotAllowed = 0x39
    secureDataVerificationFailed = 0x3A
    certificateVerificationFailedInvalidTimePeriod = 0x50
    certificateVerificationFailedInvalidSignature = 0x51
    certificateVerificationFailedInvalidChainOfTrust = 0x52
    certificateVerificationFailedInvalidType = 0x53
    certificateVerificationFailedInvalidFormat = 0x54
    certificateVerificationFailedInvalidContent = 0x55
    certificateVerificationFailedInvalidScope = 0x56
    certificateVerificationFailedInvalidCertificateRevoked = 0x57
    ownershipVerificationFailed = 0x58
    challengeCalculationFailed = 0x59
    settingAccessRightsFailed = 0x5A
    sessionKeyCreationOrDerivationFailed = 0x5B
    configurationDataUsageFailed = 0x5C
    deAuthenticationFailed = 0x5D
    uploadDownloadNotAccepted = 0x70
    transferDataSuspended = 0x71
    generalProgrammingFailure = 0x72
    wrongBlockSequenceCounter = 0x73
    requestCorrectlyReceivedResponsePending = 0x78
    subFunctionNotSupportedInActiveSession = 0x7E
    serviceNotSupportedInActiveSession = 0x7F
    rpmTooHigh = 0x81
    rpmTooLow = 0x82
    engineIsRunning = 0x83
    engineIsNotRunning = 0x84
    engineRunTimeTooLow = 0x85
    temperatureTooHigh = 0x86
    temperatureTooLow = 0x87
    vehicleSpeedTooHigh = 0x88
    vehicleSpeedTooLow = 0x89
    throttlePedalTooHigh = 0x8A
    throttlePedalTooLow = 0x8B
    transmissionRangeNotInNeutral = 0x8C
    transmissionRangeNotInGear = 0x8D
    brakeSwitchNotClosed = 0x8F
    shifterLeverNotInPark = 0x90
    torqueConverterClutchLocked = 0x91
    voltageTooHigh = 0x92
    voltageTooLow = 0x93
    resourceTemporarilyNotAvailable = 0x94
    vehicleManufacturerSpecificConditionsNotCorrectF0 = 0xF0
    vehicleManufacturerSpecificConditionsNotCorrectF1 = 0xF1
    vehicleManufacturerSpecificConditionsNotCorrectF2 = 0xF2
    vehicleManufacturerSpecificConditionsNotCorrectF3 = 0xF3
    vehicleManufacturerSpecificConditionsNotCorrectF4 = 0xF4
    vehicleManufacturerSpecificConditionsNotCorrectF5 = 0xF5
    vehicleManufacturerSpecificConditionsNotCorrectF6 = 0xF6
    vehicleManufacturerSpecificConditionsNotCorrectF7 = 0xF7
    vehicleManufacturerSpecificConditionsNotCorrectF8 = 0xF8
    vehicleManufacturerSpecificConditionsNotCorrectF9 = 0xF9
    vehicleManufacturerSpecificConditionsNotCorrectFA = 0xFA
    vehicleManufacturerSpecificConditionsNotCorrectFB = 0xFB
    vehicleManufacturerSpecificConditionsNotCorrectFC = 0xFC
    vehicleManufacturerSpecificConditionsNotCorrectFD = 0xFD
    vehicleManufacturerSpecificConditionsNotCorrectFE = 0xFE


class DSCSubFuncs(IntEnum):
    # defaultSession
    DS = 0x01
    # programmingSession
    PRGS = 0x02
    # extendedDiagnosticSession
    EXTDS = 0x03
    # safetySystemDiagnosticSession
    SSDS = 0x04


class RCSubFuncs(IntEnum):
    # start routine
    STR = 0x01
    # stop routine
    STPR = 0x02
    # request routine results
    RRR = 0x03


class CCSubFuncs(IntEnum):
    # enableRxAndTx
    ERXTX = 0x00
    # enableRxAndDisableTx
    ERXDTX = 0x01
    # disableRxAndEnableTx
    DRXETX = 0x02
    # disableRxAndTx
    DRXTX = 0x03
    # Plus vendor specific stuff...


class CDTCSSubFuncs(IntEnum):
    ON = 0x01
    OFF = 0x02
    # Plus vendor specific stuff...


class RDTCISubFuncs(IntEnum):
    # reportNumberOfDTCByStatusMask
    RNODTCBSM = 0x01
    # reportDTCByStatusMask
    RDTCBSM = 0x02
    # reportSupportedDTC
    RSUPDTC = 0x0A
    # reportFirstTestFailedDTC
    RFTFDTC = 0x0B
    # reportFirstConfirmedDTC
    RFCDTC = 0x0C
    # reportMostRecentTestFailedDTC
    RMRTFDTC = 0x0D
    # reportMostRecentConfirmedDTC
    RMRCDTC = 0x0E
    # reportMirrorMemoryDTCByStatusMask
    RMMDTCBSM = 0x0F
    # reportNumberOfMirrorMemoryDTCByStatusMask
    RNOMMDTCBSM = 0x11
    # reportNumberOfEmissionsRelatedOBDDTCByStatusMask
    RNOOBDDTCBSM = 0x12
    # reportEmissionsRelatedOBDDTCByStatusMask
    ROBDDTCBSM = 0x13
    # reportDTCFaultDetectionCounter
    RDTCFDC = 0x14
    # reportDTCWithPermanentStatus
    RDTCWPS = 0x15


class InputOutputControlParameter(IntEnum):
    # returnControlToECU
    RCTECU = 0x00
    # resetToDefault
    RTD = 0x01
    # freezeCurrentState
    FCS = 0x02
    # shortTermAdjustment
    STA = 0x03


class DTCFormatIdentifier(IntEnum):
    # ISO15031-6DTCFormat
    ISO_15031_6 = 0x00
    # ISO14229-1DTCFormat
    ISO_14229_1 = 0x01
    # SAEJ1939-73DTCFormat
    SAE_J1939_73 = 0x02
    # ISO11992-4DTCFormat
    ISO_11992_4 = 0x03


# This dictionary maps UDS services to the echo length of their responses.
# Echos in that context are values which are identical to the corresponding entry in a request.
# Therefore they can be used to match responses to requests but are not adding any new information.
# Some services (e.g. rdbi) can accept more data records in a request and response accordingly.
# In that case there can be several echos. However, as of the time of writing, multiple data
# records are not considered in the rest of the code.
# For a complete handling one might need to transfer this to a function.
UDSIsoServicesEchoLength = {
    UDSIsoServices.DiagnosticSessionControl: 1,
    UDSIsoServices.EcuReset: 1,
    UDSIsoServices.SecurityAccess: 1,
    UDSIsoServices.CommunicationControl: 1,
    UDSIsoServices.TesterPresent: 1,
    UDSIsoServices.AccessTimingParameter: 1,
    UDSIsoServices.ControlDTCSetting: 1,
    UDSIsoServices.ResponseOnEvent: 1,  # There are a number of echos but only one byte is a prefix
    UDSIsoServices.LinkControl: 1,
    UDSIsoServices.ReadDataByIdentifier: 2,
    UDSIsoServices.ReadScalingDataByIdentifier: 2,
    UDSIsoServices.ReadDataByPeriodicIdentifier: 1,  # This one is a little weird
    UDSIsoServices.DynamicallyDefineDataIdentifier: 3,
    UDSIsoServices.WriteDataByIdentifier: 2,
    # This one would require to parse the addressAndLengthFormatIdentifier field
    # UDSIsoServices.WriteMemoryByAddress: None,
    UDSIsoServices.ReadDTCInformation: 1,
    UDSIsoServices.InputOutputControlByIdentifier: 2,
    UDSIsoServices.RoutineControl: 3,
    UDSIsoServices.TransferData: 1,
}


class DataIdentifier(IntEnum):
    ActiveDiagnosticSessionDataIdentifier = 0xF186



import asyncio
import reprlib
from argparse import Namespace, BooleanOptionalAction
from binascii import unhexlify
from typing import Any

from gallia.uds.core.exception import UDSException
from gallia.uds.core.constants import UDSIsoServices
from gallia.uds.core.client import UDSRequestConfig
from gallia.uds.core.service import NegativeResponse, UDSResponse
from gallia.uds.helpers import suggests_service_not_supported
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import (
    auto_int,
    check_and_set_session,
    find_sessions,
    ParseSkips,
)


class ScanServices(UDSScanner):
    """Iterate sessions and services and find endpoints"""

    def add_parser(self) -> None:
        self.parser.add_argument(
            "--sessions",
            nargs="*",
            type=auto_int,
            default=None,
            help="Set list of sessions to be tested; all if None",
        )
        self.parser.add_argument(
            "--check-session",
            action="store_true",
            default=False,
            help="check current session with read DID",
        )
        self.parser.add_argument(
            "--scan-response-ids",
            default=False,
            action=BooleanOptionalAction,
            help="Do not scan reply flag in SID",
        )
        self.parser.add_argument(
            "--auto-reset",
            action="store_true",
            default=False,
            help="Reset ECU with UDS ECU Reset before every request",
        )
        self.parser.add_argument(
            "--payload",
            default=None,
            type=unhexlify,
            help="Payload which will be appended for each request as hex string",
        )
        self.parser.add_argument(
            "--skip",
            nargs="+",
            default={},
            type=str,
            action=ParseSkips,
            help="""
                 The service IDs to be skipped per session.
                 A session specific skip is given by <session id>:<service ids>
                 where <service ids> is a comma separated list of single ids or id ranges using a dash.
                 Examples:
                  - 0x01:0xf3
                  - 0x10-0x2f
                  - 0x01:0xf3,0x10-0x2f
                 Multiple session specific skips are separated by space.
                 """,
        )

    async def main(self, args: Namespace) -> None:
        self.ecu.max_retry = 1
        found: dict[int, dict[int, Any]] = {}

        if args.sessions is None:
            self.logger.log_info("No sessions specified, starting with session scan")
            # Only until 0x80 because the eight bit is "SuppressResponse"
            sessions = list(
                s
                for s in range(1, 0x80)
                if s not in args.skip or args.skip[s] is not None
            )
            sessions = await find_sessions(self.ecu, sessions)
            self.logger.log_summary(
                f'Found {len(sessions)} sessions: {" ".join([hex(i) for i in sessions])}'
            )
        else:
            sessions = list(
                s
                for s in args.sessions
                if s not in args.skip or args.skip[s] is not None
            )

        self.logger.log_info(f"testing sessions {sessions}")
        self.logger.log_info(f"skipping identifiers {reprlib.repr(args.skip)}")

        for session in sessions:
            self.logger.log_info(f"Switching to session 0x{session:02X}")
            try:
                resp: UDSResponse = await self.ecu.set_session(
                    session, UDSRequestConfig(tags=["preparation"])
                )
            except (UDSException, RuntimeError) as e:
                self.logger.log_warning(f"session change: {session:x} reason: {e}")
                continue
            if isinstance(resp, NegativeResponse):
                self.logger.log_warning(f"session change: {session:x} reason: {resp}")
                continue

            found[session] = {}
            self.logger.log_summary(f"scanning in session 0x{session:02X}")

            # Starts at 0x00, see first loop iteration.
            sid = -1
            while sid < 0xFF:
                sid += 1
                if sid & 0x40 and not args.scan_response_ids:
                    continue

                if session in args.skip and sid in args.skip[session]:
                    self.logger.log_info(f"0x{sid:02X}: skipped")
                    continue

                if args.check_session:
                    if not await check_and_set_session(self.ecu, session):
                        self.logger.log_error(
                            f"Aborting scan on session 0x{session:02x}; current SID was 0x{sid:0x}"
                        )
                        break

                pdu = bytes([sid]) + args.payload if args.payload else bytes([sid])

                try:
                    resp = await self.ecu.send_raw(
                        pdu, config=UDSRequestConfig(tags=["ANALYZE"])
                    )
                except asyncio.TimeoutError:
                    self.logger.log_info(f"0x{sid:02X}: timeout")
                    continue
                except Exception as e:
                    self.logger.log_info(f"0x{sid:02X}: {type(e)} occurred")
                    await self.ecu.reconnect()
                    continue

                if suggests_service_not_supported(resp):
                    self.logger.log_info(f"0x{sid:02X}: not supported [{resp}]")
                    continue

                self.logger.log_summary(
                    f"0x{sid:02X}: available in session 0x{session:02X}: {resp}"
                )
                found[session][sid] = resp

            await self.ecu.leave_session(session)

        for key, value in found.items():
            self.logger.log_summary(f"findings in session 0x{key:02X}:")
            for sid, data in value.items():
                try:
                    self.logger.log_summary(
                        f"  [0x{sid:02X}] {UDSIsoServices(sid).name}: {data}"
                    )
                except Exception:
                    self.logger.log_summary(
                        f"  [0x{sid:02X}] vendor specific sid: {data}"
                    )



from typing import Any, Optional

import pyxcp
import pyxcp.types

from gallia.penlog import Logger
from gallia.transports.base import BaseTransport


class XCPService:
    def __init__(self, transport: BaseTransport, timeout: float = 1.0) -> None:
        self.logger = Logger("xcp", flush=True)
        self.timeout = timeout
        self.transport = transport
        # This uses construct types which would result in a new
        # dependency. Let's go with Any for this attribute.
        self.byte_order: Any

    async def request(self, data: bytes, timeout: Optional[float] = None) -> bytes:
        t = timeout if timeout else self.timeout
        resp = await self.transport.request(data, t)
        header = pyxcp.types.Response.parse(resp)
        self.logger.log_info(str(header))
        if int(header.type) != 255:
            raise ValueError(
                f"Unknown response type: {header.type}, maybe no XCP packet?"
            )
        # strip header byte
        return resp[1:]

    async def connect(self) -> None:
        self.logger.log_info("XCP CONNECT")
        resp = await self.request(bytes([0xFF, 0x00]))
        tmp = pyxcp.types.ConnectResponsePartial.parse(resp)
        self.byte_order = tmp.commModeBasic.byteOrder
        tmp = pyxcp.types.ConnectResponse.parse(resp, byteOrder=self.byte_order)
        self.logger.log_info(str(tmp))
        self.logger.log_summary("XCP CONNECT -> OK")

    async def disconnect(self) -> None:
        self.logger.log_info("XCP DISCONNECT")
        resp = await self.request(bytes([0xFE, 0x00]))
        self.logger.log_info(str(resp))
        self.logger.log_summary("XCP DISCONNECT -> OK")

    async def get_status(self) -> None:
        self.logger.log_info("XCP GET_STATUS")
        resp = await self.request(bytes([0xFD]))
        tmp = pyxcp.types.GetStatusResponse.parse(resp, byteOrder=self.byte_order)
        self.logger.log_info(str(tmp))
        self.logger.log_summary("XCP GET_STATUS -> OK")

    async def get_comm_mode_info(self) -> None:
        self.logger.log_info("XCP GET_COMM_MODE_INFO")
        resp = await self.request(bytes([0xFB]))
        tmp = pyxcp.types.GetCommModeInfoResponse.parse(
            resp,
            byteOrder=self.byte_order,
        )
        self.logger.log_info(str(tmp))
        self.logger.log_summary("XCP GET_COMM_MODE_INFO -> OK")

    async def get_id(self, id_: int) -> None:
        self.logger.log_info(f"XCP GET_ID({id_})")
        resp = await self.request(bytes([0xFA, id_]))
        tmp = pyxcp.types.GetIDResponse.parse(resp, byteOrder=self.byte_order)
        self.logger.log_info(str(tmp))
        self.logger.log_summary(f"XCP GET_ID({id_}) -> OK")

    async def upload(self, length: int) -> None:
        self.logger.log_info(f"XCP GET_UPLOAD({length}")
        resp = await self.request(bytes([0xF5, length]))
        self.logger.log_info(str(resp))
        self.logger.log_summary(f"XCP GET_UPLOAD({length} -> OK")



import sys
from argparse import Namespace

from gallia.uds.core.service import NegativeResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int, check_and_set_session


class ReadMemoryByAddressScanner(UDSScanner):
    """Read memory by address"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

        self.parser.add_argument(
            "--session",
            type=auto_int,
            default=0x01,
            help="The session in which the requests are made",
        )
        self.parser.add_argument(
            "address",
            type=auto_int,
            help="The start address from which data should be read",
        )
        self.parser.add_argument(
            "length", type=auto_int, help="The number of bytes which should be read"
        )

    async def main(self, args: Namespace) -> None:
        try:
            await check_and_set_session(self.ecu, args.session)
        except Exception as e:
            self.logger.log_critical(
                f"Could not change to session: 0x{args.session:02x}: {e.__class__.__name__} {e}"
            )
            sys.exit(1)

        resp = await self.ecu.read_memory_by_address(args.address, args.length)

        if isinstance(resp, NegativeResponse):
            self.logger.log_error(resp)
        else:
            self.logger.log_summary("Positive response:")

            self.logger.log_summary(f"hex: {resp.data_record.hex()}")
            self.logger.log_summary(f"raw: {repr(resp.data_record)}")



import binascii
import sys
from argparse import Namespace

from gallia.uds.core.client import UDSRequestConfig
from gallia.uds.core.exception import UDSException
from gallia.uds.core.service import (
    NegativeResponse,
    UDSResponse,
    UDSRequest,
    RawRequest,
    RawResponse,
)
from gallia.uds.helpers import raise_for_error
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int


class SendPDU(UDSScanner):
    """A raw scanner to send a plain pdu"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

        self.parser.add_argument(
            "pdu",
            type=binascii.unhexlify,
            help="The raw pdu to send to the ECU",
        )
        self.parser.add_argument(
            "-r",
            "--max-retry",
            type=int,
            default=4,
            help="Set the uds' stack max_retry argument",
        )
        self.parser.add_argument(
            "--session",
            type=auto_int,
            default=None,
            help="Change to this session prior to sending the pdu",
        )

    async def main(self, args: Namespace) -> None:
        pdu = args.pdu
        if args.session is not None:
            resp: UDSResponse = await self.ecu.set_session(args.session)
            raise_for_error(resp)

        parsed_request = UDSRequest.parse_dynamic(pdu)

        if isinstance(parsed_request, RawRequest):
            self.logger.log_warning("Could not parse the request pdu")

        self.logger.log_info(f"Sending {parsed_request}")

        try:
            response = await self.ecu.send_raw(
                pdu, config=UDSRequestConfig(max_retry=args.max_retry)
            )
        except UDSException as e:
            self.logger.log_error(repr(e))
            sys.exit(1)

        if isinstance(response, NegativeResponse):
            self.logger.log_warning(f"Received {response}")
        else:
            if isinstance(response, RawResponse):
                self.logger.log_warning("Could not parse the response pdu")

            self.logger.log_notice(f"Received {response}")



from __future__ import annotations

# The pylint ignores are false positives in the pylint tool.

import asyncio
import gzip
import io
import os
import struct
import shlex
from asyncio import subprocess
from datetime import datetime
from functools import partial
from pathlib import Path
from socket import SocketKind  # pylint: disable=no-name-in-module
from typing import Callable, Optional, Union
from urllib.parse import urlparse

from opennetzteil import Netzteil

from gallia.utils import split_host_port
from gallia.penlog import Logger
from gallia.transports.base import TargetURI
from gallia.transports.can import ISOTPTransport, RawCANTransport


class PowerSupplyURI(TargetURI):
    def __init__(self, raw: str) -> None:
        super().__init__(raw)
        if "id" not in self.qs:
            raise ValueError("id is missing in power-supply URI")
        if "channel" not in self.qs:
            raise ValueError("channel is missing in power-supply URI")

    @property
    def id(self) -> int:
        return int(self.qs["id"][0], 0)

    @property
    def channel(self) -> Union[int, list[int]]:
        if len(ch := self.qs["channel"]) == 1:
            return int(ch[0], 0)
        return list(map(partial(int, base=0), ch))


class PowerSupply:
    def __init__(self, channel_id: Union[int, list[int]], client: Netzteil) -> None:
        self.logger = Logger("penlab.experiment", flush=True)
        self.channel_id = channel_id
        self.netzteil = client
        self.mutex = asyncio.Lock()

    @classmethod
    async def connect(cls, target: PowerSupplyURI) -> PowerSupply:
        client = await Netzteil.connect(target.location, target.id)
        return cls(target.channel, client)

    async def _power(self, op: bool) -> None:
        assert self.netzteil
        if isinstance(self.channel_id, list):
            for id_ in self.channel_id:
                await self.netzteil.set_channel(id_, op)
        elif isinstance(self.channel_id, int):
            await self.netzteil.set_channel(self.channel_id, op)

    async def power_up(self) -> None:
        self.logger.log_info("power up experiment")
        await self._power(True)

    async def power_down(self) -> None:
        self.logger.log_info("power down experiment")
        await self._power(False)

    async def power_cycle(
        self,
        sleep: float = 2.0,
        callback: Optional[Callable] = None,
    ) -> None:
        async with self.mutex:
            await self.power_down()
            await asyncio.sleep(sleep)
            await self.power_up()
            if callback:
                await callback()


class Dumpcap:
    BUFSIZE = io.DEFAULT_BUFFER_SIZE

    def __init__(
        self,
        proc: subprocess.Process,  # pylint: disable=no-member
        logger: Logger,
        artifacts_dir: Path,
        outfile: Path,
        cleanup: int = 2,
    ) -> None:
        self.proc = proc
        self.logger = logger
        self.artifacts_dir = artifacts_dir
        self.outfile = outfile
        self.cleanup = cleanup
        self.ready_event = asyncio.Event()
        self.compressor = asyncio.create_task(self._compressor())

    @classmethod
    async def start(
        cls,
        target: TargetURI,
        artifacts_dir: Optional[os.PathLike] = None,
    ) -> Dumpcap:
        logger = Logger("penlab.dumpcap", flush=True)

        if artifacts_dir:
            artifacts_dir = Path(artifacts_dir)
        elif path := os.environ.get("PENRUN_ARTIFACTS"):
            artifacts_dir = Path(path)
        else:
            raise ValueError("no artifacts dir set")

        ts = int(datetime.now().timestamp())
        if target.scheme in [ISOTPTransport.SCHEME, RawCANTransport.SCHEME]:
            outfile = artifacts_dir.joinpath(f"candump-{ts}.pcap.gz")
            cmd = cls._can_cmd(target.netloc, target.src_addr, target.dst_addr)
        else:
            outfile = artifacts_dir.joinpath(f"eth-{ts}.pcap.gz")
            cmd = await cls._eth_cmd(target.netloc)

        cmd_str = shlex.join(cmd)
        try:
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
            )
        except Exception as e:
            logger.log_error(f"Could not start dumpcap: ({e.__class__.__name__}) {e}")
            raise

        if proc.returncode:
            raise RuntimeError(f"Could not start dumpcap with: [{cmd_str}]")

        logger.log_preamble(f'Started "dumpcap": {cmd_str}')

        return cls(proc, logger, artifacts_dir, outfile)

    async def sync(self) -> None:
        await self.ready_event.wait()

    async def stop(self) -> None:
        await asyncio.sleep(self.cleanup)
        self.proc.terminate()
        await self.proc.wait()
        await self.compressor

    async def _compressor(self) -> None:
        # Gzip support in aiofiles is missing.
        # https://github.com/Tinche/aiofiles/issues/46
        ready = False
        assert self.proc.stdout
        with await asyncio.to_thread(gzip.open, self.outfile, "wb") as f:  # type: ignore
            while True:
                chunk = await self.proc.stdout.read(64 * 1024)
                if chunk == b"":
                    break
                # Dumpcap first writes the pcap header. It does this
                # once the tool is ready. We can use this is as a poor
                # man's synchronization primitive.
                if not ready:
                    ready = True
                    self.ready_event.set()
                await asyncio.to_thread(f.write, chunk)

    @staticmethod
    def _swap_bytes_16(x: int) -> int:
        return struct.unpack(">H", struct.pack("<H", x))[0]

    @staticmethod
    def _can_cmd(
        iface: str, src_addr: Optional[int], dst_addr: Optional[int]
    ) -> list[str]:
        args = ["dumpcap", "-q", "-i", iface, "-w", "-"]
        # Debug this with `dumpcap -d` or `tshark -x` to inspect the captured buffer.
        filter_ = "link[1] == 0x01"  # broadcast flag; ignore "sent by us" frames

        if src_addr is not None and dst_addr is not None:
            # TODO: Support extended CAN IDs
            if src_addr > 0x800 or dst_addr > 0x800:
                raise ValueError("Extended CAN IDs are currently not supported!")

            # Debug this with `dumpcap -d` or `tshark -x` to inspect the captured buffer.
            filter_ += (
                f"&& (link[16:2] == {Dumpcap._swap_bytes_16(src_addr):#x} "  # can_id is in little endian
                f"|| link[16:2] == {Dumpcap._swap_bytes_16(dst_addr):#x})"
            )
        args += ["-f", filter_]
        return args

    @staticmethod
    async def _eth_cmd(target_ip: str) -> list[str]:
        try:
            host, port = split_host_port(target_ip)
        except Exception as e:
            raise ValueError(f"Invalid argument for target ip: {target_ip}; {e}") from e

        if proxy := os.getenv("all_proxy"):
            url = urlparse(proxy)
            host = str(url.hostname) if url.hostname else "localhost"
            port = url.port if url.port else 1080

        loop = asyncio.get_running_loop()
        res = await loop.getaddrinfo(
            host,
            port,
            type=SocketKind.SOCK_STREAM,
        )
        # We don't do round robin, ergo we only have
        # 1 result. That's fine here. The data structure
        # of the return value of getaddrinfo() is weird,
        # but it's documented in the python docs:
        # https://docs.python.org/3/library/socket.html
        addr_tuple = res[0]
        ip, port = addr_tuple[4][0], addr_tuple[4][1]
        return [
            "dumpcap",
            "-q",
            "-i",
            "any",
            "-w",
            "-",
            "-f",
            f"host {ip} and tcp port {port}",
        ]
